@@ -1,5 +1,5 @@
 /*
-   $Id: oci8.c,v 1.33 2001/08/31 15:54:13 timbo Exp $
+   $Id: oci8.c,v 1.36 2003/03/14 17:45:27 timbo Exp $
 
    Copyright (c) 1998,1999,2000,2001  Tim Bunce
 
@@ -12,27 +12,31 @@
 
 #include "Oracle.h"
 
+#ifdef UTF8_SUPPORT
+#include <utf8.h>
+#endif
 
 #ifdef OCI_V8_SYNTAX
 
 
 DBISTATE_DECLARE;
 
-static SV *ora_long;
-static SV *ora_trunc;
-static SV *ora_cache;
-static SV *ora_cache_o;		/* for ora_open() cache override */
-
 extern int pp_exec_rset _((SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec));
 
+
 void
 dbd_init_oci(dbistate_t *dbistate)
 {
     DBIS = dbistate;
-    ora_long     = perl_get_sv("Oraperl::ora_long",      GV_ADDMULTI);
-    ora_trunc    = perl_get_sv("Oraperl::ora_trunc",     GV_ADDMULTI);
-    ora_cache    = perl_get_sv("Oraperl::ora_cache",     GV_ADDMULTI);
-    ora_cache_o  = perl_get_sv("Oraperl::ora_cache_o",   GV_ADDMULTI);
+}
+
+void
+dbd_init_oci_drh(imp_drh_t * imp_drh)
+{
+    imp_drh->ora_long    = perl_get_sv("Oraperl::ora_long",      GV_ADDMULTI);
+    imp_drh->ora_trunc   = perl_get_sv("Oraperl::ora_trunc",     GV_ADDMULTI);
+    imp_drh->ora_cache   = perl_get_sv("Oraperl::ora_cache",     GV_ADDMULTI);
+    imp_drh->ora_cache_o = perl_get_sv("Oraperl::ora_cache_o",   GV_ADDMULTI);
 }
 
 
@@ -94,21 +98,17 @@ oci_hdtype_name(ub4 hdtype)
     case OCI_HTYPE_DESCRIBE:            return "OCI_HTYPE_DESCRIBE";
     case OCI_HTYPE_SERVER:              return "OCI_HTYPE_SERVER";
     case OCI_HTYPE_SESSION:             return "OCI_HTYPE_SESSION";
-    case OCI_HTYPE_TRANS:               return "OCI_HTYPE_TRANS";
-    case OCI_HTYPE_COMPLEXOBJECT:       return "OCI_HTYPE_COMPLEXOBJECT";
-    case OCI_HTYPE_SECURITY:            return "OCI_HTYPE_SECURITY";
-    case OCI_HTYPE_SUBSCRIPTION:        return "OCI_HTYPE_SUBSCRIPTION";
     /* Descriptors */
     case OCI_DTYPE_LOB:			return "OCI_DTYPE_LOB";
     case OCI_DTYPE_SNAP:		return "OCI_DTYPE_SNAP";
     case OCI_DTYPE_RSET:		return "OCI_DTYPE_RSET";
     case OCI_DTYPE_PARAM:		return "OCI_DTYPE_PARAM";
     case OCI_DTYPE_ROWID:		return "OCI_DTYPE_ROWID";
-    case OCI_DTYPE_COMPLEXOBJECTCOMP:	return "OCI_DTYPE_COMPLEXOBJECTCOMP";
-    case OCI_DTYPE_FILE:		return "OCI_DTYPE_FILE";
-    case OCI_DTYPE_LOCATOR:		return "OCI_DTYPE_LOCATOR";
+#ifdef OCI_DTYPE_REF
+    case OCI_DTYPE_REF:			return "OCI_DTYPE_REF";
+#endif
     }
-    sv = sv_2mortal(newSViv(hdtype));
+    sv = sv_2mortal(newSViv((IV)hdtype));
     return SvPV(sv,na);
 }
 
@@ -138,7 +138,7 @@ oci_error_get(OCIError *errhp, sword status, char *what, SV *errstr, int debug)
 	&& recno < 100
     ) {
 	if (debug >= 4 || recno>1/*XXX temp*/)
-	    fprintf(DBILOGFP, "    OCIErrorGet after %s (er%ld:%s): %d, %ld: %s\n",
+	    PerlIO_printf(DBILOGFP, "    OCIErrorGet after %s (er%ld:%s): %d, %ld: %s\n",
 		what, (long)recno,
 		    (eg_status==OCI_SUCCESS) ? "ok" : oci_status_name(eg_status),
 		    status, (long)eg_errcode, errbuf);
@@ -226,6 +226,8 @@ oci_db_handle(imp_dbh_t *imp_dbh, int handle_type, int flags)
      case OCI_HTYPE_SESSION:	return imp_dbh->authp;
      }
      croak("Can't get OCI handle type %d from DBI database handle", handle_type);
+     /* satisfy compiler warning, even though croak will never return */
+     return 0;
 }
 
 void *
@@ -239,6 +241,8 @@ oci_st_handle(imp_sth_t *imp_sth, int handle_type, int flags)
      case OCI_HTYPE_STMT:	return imp_sth->stmhp;
      }
      croak("Can't get OCI handle type %d from DBI statement handle", handle_type);
+     /* satisfy compiler warning, even though croak will never return */
+     return 0;
 }
 
 
@@ -279,8 +283,11 @@ dbd_st_prepare(sth, imp_sth, statement, attribs)
     imp_sth->auto_lob = 1;
     if (attribs) {
 	SV **svp;
+	long tmp;
 	DBD_ATTRIB_GET_IV(  attribs, "ora_parse_lang", 14, svp, oparse_lng);
-	DBD_ATTRIB_GET_IV(  attribs, "ora_auto_lob",   12, svp, imp_sth->auto_lob);
+	/* JLU avoid warning */
+	DBD_ATTRIB_GET_IV(  attribs, "ora_auto_lob",   12, svp, tmp);
+	imp_sth->auto_lob = (U16)tmp;
 	/* ora_check_sql only works for selects owing to Oracle behaviour */
 	DBD_ATTRIB_GET_IV(  attribs, "ora_check_sql",  13, svp, ora_check_sql);
     }
@@ -313,7 +320,7 @@ dbd_st_prepare(sth, imp_sth, statement, attribs)
 
     OCIAttrGet_stmhp_stat(imp_sth, &imp_sth->stmt_type, 0, OCI_ATTR_STMT_TYPE, status);
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP, "    dbd_st_prepare'd sql %s\n",
+	PerlIO_printf(DBILOGFP, "    dbd_st_prepare'd sql %s\n",
 		oci_stmt_type_name(imp_sth->stmt_type));
 
     DBIc_IMPSET_on(imp_sth);
@@ -325,8 +332,11 @@ dbd_st_prepare(sth, imp_sth, statement, attribs)
     else {
       /* set initial cache size by memory */
       ub4 cache_mem;
-      if      (SvOK(ora_cache_o)) cache_mem = -SvIV(ora_cache_o);
-      else if (SvOK(ora_cache))   cache_mem = -SvIV(ora_cache);
+      D_imp_dbh_from_sth ;  
+      D_imp_drh_from_dbh ;
+
+      if      (SvOK(imp_drh -> ora_cache_o)) cache_mem = -SvIV(imp_drh -> ora_cache_o);
+      else if (SvOK(imp_drh -> ora_cache))   cache_mem = -SvIV(imp_drh -> ora_cache);
       else                        cache_mem = -imp_dbh->RowCacheSize;
       if (cache_mem <= 0)
 	cache_mem = 10 * 1460;
@@ -370,7 +380,7 @@ dbd_phs_in(dvoid *octxp, OCIBind *bindp, ub4 iter, ub4 index,
     *indpp  = &phs->indp;
     *piecep = OCI_ONE_PIECE;
     if (DBIS->debug >= 3)
- 	fprintf(DBILOGFP, "       in  '%s' [%ld,%ld]: len %2ld, ind %d%s\n",
+ 	PerlIO_printf(DBILOGFP, "       in  '%s' [%ld,%ld]: len %2ld, ind %d%s\n",
 		phs->name, ul_t(iter), ul_t(index), ul_t(phs->alen), phs->indp,
 		(phs->desc_h) ? " via descriptor" : "");
     if (index > 0 || iter > 0)
@@ -440,19 +450,19 @@ dbd_phs_out(dvoid *octxp, OCIBind *bindp,
 	SV *sv = phs->sv;
 	if (SvTYPE(sv) == SVt_RV && SvTYPE(SvRV(sv)) == SVt_PVAV) {
 	    if (index > 0)	/* finish-up handling previous element */
-		dbd_phs_avsv_complete(phs, index-1, DBIS->debug);
-	    sv = *av_fetch((AV*)SvRV(sv), index, 1);
+		dbd_phs_avsv_complete(phs, (I32)index-1, DBIS->debug);
+	    sv = *av_fetch((AV*)SvRV(sv), (IV)index, 1);
 	    if (!SvOK(sv))
 		sv_setpv(sv,"");
 	}
-	*bufpp = SvGROW(sv, ((phs->maxlen < 28) ? 28 : phs->maxlen)+1/*for null*/);
+	*bufpp = SvGROW(sv, (size_t)(((phs->maxlen < 28) ? 28 : phs->maxlen)+1)/*for null*/);
 	phs->alen = SvLEN(sv);	/* max buffer size now, actual data len later */
     }
     *alenpp = &phs->alen;
     *indpp  = &phs->indp;
     *rcodepp= &phs->arcode;
     if (DBIS->debug >= 3)
- 	fprintf(DBILOGFP, "       out '%s' [%ld,%ld]: alen %2ld, piece %d%s\n",
+ 	PerlIO_printf(DBILOGFP, "       out '%s' [%ld,%ld]: alen %2ld, piece %d%s\n",
 		phs->name, ul_t(iter), ul_t(index), ul_t(phs->alen), *piecep,
 		(phs->desc_h) ? " via descriptor" : "");
     if (iter > 0)
@@ -462,18 +472,92 @@ dbd_phs_out(dvoid *octxp, OCIBind *bindp,
 }
 
 
+#ifdef UTF8_SUPPORT
+/* How many bytes are n utf8 chars in buffer */
+static ub4
+   ora_utf8_to_bytes (ub1 *buffer, ub4 chars_wanted, ub4 max_bytes)
+{
+    ub4 i = 0;
+    while (i < max_bytes && (chars_wanted-- > 0)) {
+	i += UTF8SKIP(&buffer[i]);
+    }
+    return (i < max_bytes)? i : max_bytes;
+}
+
+/* Given the 5.6.0 implementation of utf8 handling in perl,
+ * avoid setting the UTF8 flag as much as possible. Almost
+ * every binary operator in Perl will do conversions when
+ * strings marked as UTF8 are involved.
+ * Maybe setting the flag should be default in Japan or
+ * Europe? Deduce that from NLS_LANG? Possibly...
+ */
+#define DBD_SET_UTF8(sv)   (cs_is_utf8? set_utf8(sv): 0)
+static int 
+   set_utf8(SV *sv) {
+    ub1 *c;
+    for (c = SvPVX(sv); c < (ub1*)SvEND(sv); c++) {
+	if (*c & 0x80) {
+	    SvUTF8_on(sv);
+	    return 1;
+	}
+    }
+    return 0;
+}
+#endif
+
 static int	/* LONG and LONG RAW */
 fetch_func_varfield(SV *sth, imp_fbh_t *fbh, SV *dest_sv)
 {
+    D_imp_sth(sth);
+    D_imp_dbh_from_sth ;  
+    D_imp_drh_from_dbh ;
     fb_ary_t *fb_ary = fbh->fb_ary;
     char *p = (char*)&fb_ary->abuf[0];
     ub4 datalen = *(ub4*)p;     /* XXX alignment ? */
     p += 4;
-    sv_setpvn(dest_sv, p, (STRLEN)datalen);
+
+#ifdef UTF8_SUPPORT
+    if (cs_is_utf8 && fbh->ftype == 94) {
+	if (datalen > imp_sth->long_readlen) {
+	    ub4 bytelen = ora_utf8_to_bytes(p, (ub4)imp_sth->long_readlen, datalen);
+
+	    if (bytelen < datalen) {	/* will be truncated */
+		int oraperl = DBIc_COMPAT(imp_sth);
+		if (DBIc_has(imp_sth,DBIcf_LongTruncOk) 
+		      || (oraperl && SvIV(imp_drh->ora_trunc))) {
+		    /* user says truncation is ok */
+		    /* Oraperl recorded the truncation in ora_errno so we	*/
+		    /* so also but only for Oraperl mode handles.		*/
+		    if (oraperl) sv_setiv(DBIc_ERR(imp_sth), 1406);
+		} else {
+		    char buf[300];
+		    sprintf(buf,"fetching field %d of %d. LONG value truncated from %ld to %ld. %s",
+			    fbh->field_num+1, DBIc_NUM_FIELDS(imp_sth), datalen, bytelen,
+			    "DBI attribute LongReadLen too small and/or LongTruncOk not set");
+		    oci_error(sth, NULL, OCI_ERROR, buf);
+		    sv_setiv(DBIc_ERR(imp_sth), (IV)24345); /* appropriate ORA error number */
+		    (void)SvOK_off(dest_sv);
+		    return 0;
+		}
+
+		if (DBIS->debug >= 3)
+		    PerlIO_printf(DBILOGFP, "       fetching field %d of %d. LONG value truncated from %ld to %ld.\n",
+			    fbh->field_num+1, DBIc_NUM_FIELDS(imp_sth), datalen, bytelen);
+		datalen = bytelen;
+	    }
+	}
+	sv_setpvn(dest_sv, p, (STRLEN)datalen);
+	DBD_SET_UTF8(dest_sv);
+    } else {
+#else
+    {
+#endif
+	sv_setpvn(dest_sv, p, (STRLEN)datalen);
+    }
+
     return 1;
 }
 
-
 static int
 fetch_func_nty(SV *sth, imp_fbh_t *fbh, SV *dest_sv)
 {
@@ -500,7 +584,7 @@ pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec)
         HV *init_attr = newHV();
 	int count;
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "       bind %s - allocating new sth...\n", phs->name);
+	    PerlIO_printf(DBILOGFP, "       bind %s - allocating new sth...\n", phs->name);
 	ENTER;
 	PUSHMARK(SP);
 	XPUSHs(sv_2mortal(newRV(DBIc_MY_H(imp_dbh))));
@@ -515,7 +599,7 @@ pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec)
 	PUTBACK;
 	LEAVE;
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "       bind %s - allocated %s...\n",
+	    PerlIO_printf(DBILOGFP, "       bind %s - allocated %s...\n",
 		phs->name, neatsvpv(phs->sv, 0));
 
     }
@@ -525,7 +609,7 @@ pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec)
 	D_impdata(imp_sth_csr, imp_sth_t, sth_csr);
 
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "       bind %s - initialising new %s...\n",
+	    PerlIO_printf(DBILOGFP, "       bind %s - initialising new %s...\n",
 		phs->name, neatsvpv(sth_csr,0));
 
 #ifdef OCI_V8_SYNTAX
@@ -607,6 +691,97 @@ dbd_rebind_ph_lob(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 }
 
 
+#ifdef UTF8_SUPPORT
+ub4
+ora_blob_read_mb_piece(SV *sth, imp_sth_t *imp_sth, imp_fbh_t *fbh, 
+  SV *dest_sv, long offset, long len, long destoffset)
+{
+    ub4 loblen = 0;
+    ub4 buflen;
+    ub4 amtp = 0;
+    ub4 byte_destoffset = 0;
+    OCILobLocator *lobl = (OCILobLocator*)fbh->desc_h;
+    sword ftype = fbh->ftype;
+    sword status;
+
+    /*
+     * We assume our caller has already done the
+     * equivalent of the following:
+     *		(void)SvUPGRADE(dest_sv, SVt_PV);
+     */
+
+    if (ftype != 112) {
+	oci_error(sth, imp_sth->errhp, OCI_ERROR,
+	"blob_read not currently supported for non-CLOB types with OCI 8 "
+	"(but with OCI 8 you can set $dbh->{LongReadLen} to the length you need,"
+	"so you don't need to call blob_read at all)");
+	(void)SvOK_off(dest_sv);	/* signal error */
+	return 0;
+    }
+
+    OCILobGetLength_log_stat(imp_sth->svchp, imp_sth->errhp, 
+			     lobl, &loblen, status);
+    if (status != OCI_SUCCESS) {
+	oci_error(sth, imp_sth->errhp, status, "OCILobGetLength");
+	(void)SvOK_off(dest_sv);	/* signal error */
+	return 0;
+    }
+
+    loblen -= offset;   /* only count from offset onwards */
+    amtp = (loblen > len) ? len : loblen;
+    buflen = 4 * amtp;
+
+    byte_destoffset = ora_utf8_to_bytes((ub1 *)(SvPVX(dest_sv)), 
+				    (ub4)destoffset, SvCUR(dest_sv));
+    
+    if (loblen > 0) {
+      ub1 *dest_bufp;
+      ub1 *buffer;
+
+      New(42, buffer, buflen, ub1);
+
+      OCILobRead_log_stat(imp_sth->svchp, imp_sth->errhp, lobl,
+			  &amtp, (ub4)1 + offset, buffer, buflen,
+			  0, 0, (ub2)0, (ub1)SQLCS_IMPLICIT, status);
+      if (dbis->debug >= 3)
+	PerlIO_printf(DBILOGFP, "       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
+		fbh->field_num+1, oci_status_name(status), loblen, 
+		imp_sth->long_readlen, buflen, amtp);
+      if (status != OCI_SUCCESS) {
+	oci_error(sth, imp_sth->errhp, status, "OCILobRead");
+	(void)SvOK_off(dest_sv);	/* signal error */
+	return 0;
+      }
+
+      amtp = ora_utf8_to_bytes(buffer, len, amtp);
+      SvGROW(dest_sv, byte_destoffset + amtp + 1);
+      dest_bufp = (ub1 *)(SvPVX(dest_sv));
+      dest_bufp += byte_destoffset;
+      memcpy(dest_bufp, buffer, amtp);
+      Safefree(buffer);
+    }
+    else {
+      assert(amtp == 0);
+      SvGROW(dest_sv, byte_destoffset + 1);
+      if (dbis->debug >= 3)
+	PerlIO_printf(DBILOGFP,
+		"       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
+		fbh->field_num+1, "SKIPPED", loblen, imp_sth->long_readlen, buflen, amtp);
+    }
+
+    if (dbis->debug >= 3)
+      PerlIO_printf(DBILOGFP, "    blob_read field %d, ftype %d, offset %ld, len %ld, destoffset %ld, retlen %ld\n",
+	      fbh->field_num+1, ftype, offset, len, destoffset, amtp);
+    
+    SvCUR_set(dest_sv, byte_destoffset+amtp);
+    *SvEND(dest_sv) = '\0'; /* consistent with perl sv_setpvn etc	*/
+    SvPOK_on(dest_sv);
+    DBD_SET_UTF8(dest_sv);
+
+    return 1;
+}
+#endif /* ifdef UTF8_SUPPORT */
+
 ub4
 ora_blob_read_piece(SV *sth, imp_sth_t *imp_sth, imp_fbh_t *fbh, SV *dest_sv,
 		    long offset, long len, long destoffset)
@@ -649,10 +824,10 @@ ora_blob_read_piece(SV *sth, imp_sth_t *imp_sth, imp_fbh_t *fbh, SV *dest_sv,
 	bufp += destoffset;
 
 	OCILobRead_log_stat(imp_sth->svchp, imp_sth->errhp, lobl,
-	    &amtp, 1 + offset, bufp, buflen,
-			    0, 0, 0, SQLCS_IMPLICIT, status);
+	    &amtp, (ub4)1 + offset, bufp, buflen,
+			    0, 0, (ub2)0, (ub1)SQLCS_IMPLICIT, status);
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
 		fbh->field_num+1, oci_status_name(status), ul_t(loblen),
 		imp_sth->long_readlen, ul_t(buflen), ul_t(amtp));
@@ -665,7 +840,7 @@ ora_blob_read_piece(SV *sth, imp_sth_t *imp_sth, imp_fbh_t *fbh, SV *dest_sv,
     else {
 	assert(amtp == 0);
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
 		fbh->field_num+1, "SKIPPED", ul_t(loblen),
 		imp_sth->long_readlen, ul_t(buflen), ul_t(amtp));
@@ -708,7 +883,10 @@ fetch_func_autolob(SV *sth, imp_fbh_t *fbh, SV *dest_sv)
 
     if (loblen > imp_sth->long_readlen) {	/* LOB will be truncated */
 	int oraperl = DBIc_COMPAT(imp_sth);
-	if (DBIc_has(imp_sth,DBIcf_LongTruncOk) || (oraperl && SvIV(ora_trunc))) {
+	D_imp_dbh_from_sth ;  
+	D_imp_drh_from_dbh ;
+
+	if (DBIc_has(imp_sth,DBIcf_LongTruncOk) || (oraperl && SvIV(imp_drh -> ora_trunc))) {
 	    /* user says truncation is ok */
 	    /* Oraperl recorded the truncation in ora_errno so we	*/
 	    /* so also but only for Oraperl mode handles.		*/
@@ -732,35 +910,93 @@ fetch_func_autolob(SV *sth, imp_fbh_t *fbh, SV *dest_sv)
 	buflen = amtp * 4;  /* XXX bit of a hack, efective but wasteful */
     }
     else buflen = amtp;
+
     (void)SvUPGRADE(dest_sv, SVt_PV);
+
     SvGROW(dest_sv, buflen+1);
 
     if (loblen > 0) {
+#ifdef UTF8_SUPPORT
+	if (cs_is_utf8 && fbh->ftype == 112) {
+	    ub4 alloclen = buflen << 2;
+	    ub4 chars_to_read = amtp;
+	    char *buffer;
+	    New(42, buffer, alloclen, char);
+	    OCILobRead_log_stat(imp_sth->svchp, imp_sth->errhp, lobloc,
+				&amtp, (ub4)1, buffer, alloclen, 
+				0, 0, (ub2)0, (ub1)SQLCS_IMPLICIT, status);
+	    
+	    if (DBIS->debug >= 3) {
+		PerlIO_printf(DBILOGFP, "       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
+			      fbh->field_num+1, oci_status_name(status), loblen, 
+			      imp_sth->long_readlen, alloclen, amtp);
+	    }
+	    if (status != OCI_SUCCESS) {
+		oci_error(sth, imp_sth->errhp, status, "OCILobRead");
+		(void)SvOK_off(dest_sv);
+		return 0;
+	    }
+	           
+	    SvGROW(dest_sv, amtp+1);
+	    memcpy(SvPVX(dest_sv), buffer, amtp);
+	    Safefree(buffer);
+	    
+	    /* tell perl what we've put in its dest_sv */
+	    SvCUR(dest_sv) = amtp;
+	    *SvEND(dest_sv) = '\0';
+	    DBD_SET_UTF8(dest_sv);
+	} 
+	else
+#endif /* ifdef UTF8_SUPPORT */
+	{
+	    SvGROW(dest_sv, buflen+1);
+	    if (fbh->dbtype == 114) {
+	    OCILobFileOpen_log_stat(imp_sth->svchp, imp_sth->errhp, lobloc,
+				    (ub1)OCI_FILE_READONLY, status);
+	    if (status != OCI_SUCCESS) {
+		oci_error(sth, imp_sth->errhp, status, "OCILobFileOpen");
+		(void)SvOK_off(dest_sv);
+		return 0;
+	    }
+	}
+
 	OCILobRead_log_stat(imp_sth->svchp, imp_sth->errhp, lobloc,
-	    &amtp, 1, SvPVX(dest_sv), buflen, 0, 0, 0, SQLCS_IMPLICIT, status);
+	    &amtp, (ub4)1, SvPVX(dest_sv), buflen,
+	    0, 0, (ub2)0, (ub1)SQLCS_IMPLICIT, status);
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       OCILobRead field %d %s: LOBlen %ldc, LongReadLen %ldc, BufLen %ldb, Got %ldc\n",
 		fbh->field_num+1, oci_status_name(status), ul_t(loblen),
 		imp_sth->long_readlen, ul_t(buflen), ul_t(amtp));
+	if (fbh->dbtype == 114) {
+	    OCILobFileClose_log_stat(imp_sth->svchp, imp_sth->errhp,
+		lobloc, status);
+	}
 	if (status != OCI_SUCCESS) {
 	    oci_error(sth, imp_sth->errhp, status, "OCILobRead");
 	    (void)SvOK_off(dest_sv);
 	    return 0;
 	}
+	
+	/* tell perl what we've put in its dest_sv */
+	SvCUR(dest_sv) = amtp;
+	*SvEND(dest_sv) = '\0';
+	}
+	
     }
     else {
 	assert(amtp == 0);
+	/* tell perl what we've put in its dest_sv */
+	SvGROW(dest_sv, buflen+1);
+	SvCUR(dest_sv) = amtp;
+	*SvEND(dest_sv) = '\0';
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       OCILobRead field %d %s: LOBlen %ld, LongReadLen %ld, BufLen %ld, Got %ld\n",
 		fbh->field_num+1, "SKIPPED", ul_t(loblen),
 		imp_sth->long_readlen, ul_t(buflen), ul_t(amtp));
     }
 
-    /* tell perl what we've put in its dest_sv */
-    SvCUR(dest_sv) = amtp;
-    *SvEND(dest_sv) = '\0';
     SvPOK_on(dest_sv);
 
     return 1;
@@ -779,7 +1015,7 @@ static void
 fbh_setup_getrefpv(imp_fbh_t *fbh, int desc_t, char *bless)
 {
     if (DBIS->debug >= 2)
-	fprintf(DBILOGFP,
+	PerlIO_printf(DBILOGFP,
 	    "    col %d: otype %d, desctype %d, %s", fbh->field_num, fbh->dbtype, desc_t, bless);
     fbh->ftype  = fbh->dbtype;
     fbh->disize = fbh->dbsize;
@@ -794,6 +1030,7 @@ int
 dbd_describe(SV *h, imp_sth_t *imp_sth)
 {
     D_imp_dbh_from_sth;
+	D_imp_drh_from_dbh ;
     I32	long_readlen;
     ub4 num_fields;
     int has_longs = 0;
@@ -808,21 +1045,21 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
     /* ora_trunc is checked at fetch time */
     /* long_readlen:	length for long/longraw (if >0), else 80 (ora app dflt)	*/
     /* Ought to be for COMPAT mode only but was relaxed before LongReadLen existed */
-    long_readlen = (SvOK(ora_long) && SvIV(ora_long)>0)
-				? SvIV(ora_long) : DBIc_LongReadLen(imp_sth);
+    long_readlen = (SvOK(imp_drh -> ora_long) && SvIV(imp_drh -> ora_long)>0)
+				? SvIV(imp_drh -> ora_long) : DBIc_LongReadLen(imp_sth);
     if (long_readlen < 0)		/* trap any sillyness */
 	long_readlen = 80;		/* typical oracle app default	*/
 
-    if (imp_sth->stmt_type != OCI_STMT_SELECT) {
+    if (imp_sth->stmt_type != OCI_STMT_SELECT) { /* XXX DISABLED, see num_fields test below */
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "    dbd_describe skipped for %s\n",
+	    PerlIO_printf(DBILOGFP, "    dbd_describe skipped for %s\n",
 		oci_stmt_type_name(imp_sth->stmt_type));
 	/* imp_sth memory was cleared when created so no setup required here	*/
 	return 1;
     }
 
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP, "    dbd_describe %s (%s, lb %ld)...\n",
+	PerlIO_printf(DBILOGFP, "    dbd_describe %s (%s, lb %ld)...\n",
 	    oci_stmt_type_name(imp_sth->stmt_type),
 	    DBIc_ACTIVE(imp_sth) ? "implicit" : "EXPLICIT", (long)long_readlen);
 
@@ -843,6 +1080,14 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 	oci_error(h, imp_sth->errhp, status, "OCIAttrGet OCI_ATTR_PARAM_COUNT");
 	return 0;
     }
+    if (num_fields == 0) {
+	if (DBIS->debug >= 3)
+	    PerlIO_printf(DBILOGFP, "    dbd_describe skipped for %s (no fields returned)\n",
+		oci_stmt_type_name(imp_sth->stmt_type));
+	/* imp_sth memory was cleared when created so no setup required here	*/
+	return 1;
+    }
+
     DBIc_NUM_FIELDS(imp_sth) = num_fields;
     Newz(42, imp_sth->fbh, num_fields, imp_fbh_t);
 
@@ -883,6 +1128,8 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 	switch (fbh->dbtype) {
 	/*	the simple types	*/
 	case   1:				/* VARCHAR2	*/
+		avg_width = fbh->dbsize / 2;
+		/* FALLTHRU */
 	case  96:				/* CHAR		*/
 		fbh->disize = fbh->dbsize;
 		fbh->prec   = fbh->disize;
@@ -901,9 +1148,15 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 		/* actually dependent on NLS default date format*/
 		fbh->disize = 75;	/* a generous default	*/
 		fbh->prec   = fbh->disize;
+		avg_width = 8;	/* size in SQL*Net packet  */
 		break;
 
 	case   8:				/* LONG		*/
+#ifdef UTF8_SUPPORT
+	    if (cs_is_utf8)
+		fbh->disize = long_readlen * 4;
+	    else
+#endif
 		fbh->disize = long_readlen;
 		fbh->dbsize = (fbh->disize>65535) ? 65535 : fbh->disize;
 		fbh->ftype  = 94; /* VAR form */
@@ -926,6 +1179,7 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 
 	case 112:				/* CLOB		*/
 	case 113:				/* BLOB		*/
+	case 114:				/* BFILE	*/
 		fbh->ftype  = fbh->dbtype;
 		fbh->disize = fbh->dbsize;
 		fbh->fetch_func = (imp_sth->auto_lob)
@@ -941,18 +1195,23 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 		break;
 #endif
 
+	case 188: 	           /* TIMESTAMP WITH TIME ZONE	*/
+		/* actually dependent on NLS default date format*/
+		fbh->disize = 75;       /* a generous default   */
+		break;
+
 	default:
 		/* XXX unhandled type may lead to errors or worse */
 		fbh->ftype  = fbh->dbtype;
 		fbh->disize = fbh->dbsize;
 		p = "Field %d has an Oracle type (%d) which is not explicitly supported%s";
 		if (DBIS->debug >= 1)
-		    fprintf(DBILOGFP, p, i, fbh->dbtype, "\n");
+		    PerlIO_printf(DBILOGFP, p, i, fbh->dbtype, "\n");
 		if (dowarn)
 		    warn(p, i, fbh->dbtype, "");
 		break;
 	}
-	if (fbh->ftype == 5)
+	if (fbh->ftype == 5)	/* XXX need to handle wide chars somehow */
 	    fbh->disize += 1;	/* allow for null terminator */
 
 	/* dbsize can be zero for 'select NULL ...'			*/
@@ -969,11 +1228,11 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
     /* --- Setup the row cache for this query --- */
 
     /* number of rows to cache	*/
-    if      (SvOK(ora_cache_o)) imp_sth->cache_rows = SvIV(ora_cache_o);
-    else if (SvOK(ora_cache))   imp_sth->cache_rows = SvIV(ora_cache);
+    if      (SvOK(imp_drh->ora_cache_o)) imp_sth->cache_rows = SvIV(imp_drh->ora_cache_o);
+    else if (SvOK(imp_drh->ora_cache))   imp_sth->cache_rows = SvIV(imp_drh->ora_cache);
     else                        imp_sth->cache_rows = imp_dbh->RowCacheSize;
     if (imp_sth->cache_rows >= 0) {	/* set cache size by row count	*/
-	ub4 cache_rows = calc_cache_rows(num_fields,
+	ub4 cache_rows = calc_cache_rows((int)num_fields,
 				est_width, imp_sth->cache_rows, has_longs);
 	imp_sth->cache_rows = cache_rows;	/* record updated value */
 	OCIAttrSet_log_stat(imp_sth->stmhp, OCI_HTYPE_STMT,
@@ -1017,10 +1276,11 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
 	fbh->fb_ary = fb_ary_alloc(define_len, 1);
 	fb_ary = fbh->fb_ary;
 
-	OCIDefineByPos_log_stat(imp_sth->stmhp, &fbh->defnp, imp_sth->errhp, (ub4) i,
+	OCIDefineByPos_log_stat(imp_sth->stmhp, &fbh->defnp,
+	    imp_sth->errhp, (ub4) i,
 	    (fbh->desc_h) ? (dvoid*)&fbh->desc_h : (dvoid*)fb_ary->abuf,
 	    (fbh->desc_h) ?                   -1 :         define_len,
-	    fbh->ftype,
+	    (ub2)fbh->ftype,
 	    fb_ary->aindp,
 	    (ftype==94||ftype==95) ? NULL : fb_ary->arlen,
 	    fb_ary->arcode, OCI_DEFAULT, status);
@@ -1037,7 +1297,7 @@ dbd_describe(SV *h, imp_sth_t *imp_sth)
     }
 
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP,
+	PerlIO_printf(DBILOGFP,
 	"    dbd_describe'd %d columns (row bytes: %d max, %d est avg, cache: %d)\n",
 	(int)num_fields, imp_sth->t_dbsize, imp_sth->est_width, imp_sth->cache_rows);
 
@@ -1070,9 +1330,9 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
     }
     else {
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "    dbd_st_fetch %d fields...\n", DBIc_NUM_FIELDS(imp_sth));
-	OCIStmtFetch_log_stat(imp_sth->stmhp, imp_sth->errhp, 1, OCI_FETCH_NEXT,
-	 OCI_DEFAULT, status);
+	    PerlIO_printf(DBILOGFP, "    dbd_st_fetch %d fields...\n", DBIc_NUM_FIELDS(imp_sth));
+	OCIStmtFetch_log_stat(imp_sth->stmhp, imp_sth->errhp,
+		1, (ub2)OCI_FETCH_NEXT, OCI_DEFAULT, status);
     }
 
     if (status != OCI_SUCCESS) {
@@ -1081,7 +1341,7 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
 	    dTHR; 			/* for DBIc_ACTIVE_off	*/
 	    DBIc_ACTIVE_off(imp_sth);	/* eg finish		*/
 	    if (DBIS->debug >= 3)
-		fprintf(DBILOGFP, "    dbd_st_fetch no-more-data\n");
+		PerlIO_printf(DBILOGFP, "    dbd_st_fetch no-more-data\n");
 	    return Nullav;
 	}
 	if (status != OCI_SUCCESS_WITH_INFO) {
@@ -1097,7 +1357,7 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
     av = DBIS->get_fbav(imp_sth);
 
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP, "    dbd_st_fetch %d fields %s\n",
+	PerlIO_printf(DBILOGFP, "    dbd_st_fetch %d fields %s\n",
 			num_fields, oci_status_name(status));
 
     ChopBlanks = DBIc_has(imp_sth, DBIcf_ChopBlanks);
@@ -1113,7 +1373,10 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
 	    && ora_dbtype_is_long(fbh->dbtype)/* field is a LONG	*/
 	) {
 	    int oraperl = DBIc_COMPAT(imp_sth);
-	    if (DBIc_has(imp_sth,DBIcf_LongTruncOk) || (oraperl && SvIV(ora_trunc))) {
+	    D_imp_dbh_from_sth ;  
+	    D_imp_drh_from_dbh ;
+
+	    if (DBIc_has(imp_sth,DBIcf_LongTruncOk) || (oraperl && SvIV(imp_drh -> ora_trunc))) {
 		/* user says truncation is ok */
 		/* Oraperl recorded the truncation in ora_errno so we	*/
 		/* so also but only for Oraperl mode handles.		*/
@@ -1138,6 +1401,9 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
 			--datalen;
 		}
 		sv_setpvn(sv, p, (STRLEN)datalen);
+#ifdef UTF8_SUPPORT
+		DBD_SET_UTF8(sv);
+#endif
 	    }
 
 	} else if (rc == 1405) {	/* field is null - return undef	*/
@@ -1153,6 +1419,9 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
 		    /* but it'll only be accessible via prior bind_column()	*/
 		    sv_setpvn(sv, (char*)&fb_ary->abuf[0],
 			  fb_ary->arlen[0]);
+#ifdef UTF8_SUPPORT
+		    DBD_SET_UTF8(sv);
+#endif
 		}
 		if (ora_dbtype_is_long(fbh->dbtype))	/* double check */
 		    hint = ", LongReadLen too small and/or LongTruncOk not set";
@@ -1169,7 +1438,7 @@ dbd_st_fetch(SV *sth, imp_sth_t *imp_sth)
 	}
 
 	if (DBIS->debug >= 5)
-	    fprintf(DBILOGFP, "        %d (rc=%d): %s\n",
+	    PerlIO_printf(DBILOGFP, "        %d (rc=%d): %s\n",
 		i, rc, neatsvpv(sv,0));
     }
 
@@ -1278,7 +1547,7 @@ find_ident_after(char *src, char *after, STRLEN *len, int copy)
 	else if (isALPHA(*src)) {
 	    if (seen_key) {
 		char *start = src;
-		while(*src && (isALNUM(*src) || *src=='.'))
+		while(*src && (isALNUM(*src) || *src=='.' || *src=='$'))
 		    ++src;
 		*len = src - start;
 		if (copy) {
@@ -1338,6 +1607,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
     lob_refetch_t *lr = NULL;
     STRLEN tablename_len;
     char *tablename;
+    char new_tablename[100];
 
     switch (imp_sth->stmt_type) {
     case OCI_STMT_UPDATE:
@@ -1357,13 +1627,32 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 		"Unable to parse table name for LOB refetch");
 
     OCIHandleAlloc_ok(imp_sth->envhp, &dschp, OCI_HTYPE_DESCRIBE, status);
+    /* BEGIN NEW */
     OCIDescribeAny_log_stat(imp_sth->svchp, errhp, tablename, strlen(tablename),
-		OCI_OTYPE_NAME, 1, OCI_PTYPE_TABLE, dschp, status);
+		(ub1)OCI_OTYPE_NAME, (ub1)1, (ub1)OCI_PTYPE_SYN, dschp, status);
+    if (status == OCI_SUCCESS) { /* There is a synonym, get the schema */
+      char *syn_schema=NULL,  *syn_name=NULL;
+      OCIAttrGet_log_stat(dschp,  OCI_HTYPE_DESCRIBE,
+				  &parmhp, 0, OCI_ATTR_PARAM, errhp, status);
+      OCIAttrGet_log_stat(parmhp, OCI_DTYPE_PARAM,
+			  &syn_schema, 0, OCI_ATTR_SCHEMA_NAME, errhp, status);
+      OCIAttrGet_log_stat(parmhp, OCI_DTYPE_PARAM,
+			  &syn_name, 0, OCI_ATTR_OBJ_NAME, errhp, status);
+      strcpy(new_tablename, syn_schema);
+      strcat(new_tablename, ".");
+      strcat(new_tablename, syn_name);
+      tablename=new_tablename;
+      if (DBIS->debug >= 3)
+	PerlIO_printf(DBILOGFP, "       lob refetch synonym, schema=%s, name=%s, new tablename=%s\n", syn_schema, syn_name, tablename);
+    }
+    /* END NEW */
+    OCIDescribeAny_log_stat(imp_sth->svchp, errhp, tablename, strlen(tablename),
+	(ub1)OCI_OTYPE_NAME, (ub1)1, (ub1)OCI_PTYPE_TABLE, dschp, status);
     if (status != OCI_SUCCESS) {
       /* XXX this OCI_PTYPE_TABLE->OCI_PTYPE_VIEW fallback should actually be	*/
       /* a loop that includes synonyms etc */
       OCIDescribeAny_log_stat(imp_sth->svchp, errhp, tablename, strlen(tablename),
-		OCI_OTYPE_NAME, 1, OCI_PTYPE_VIEW, dschp, status);
+	    (ub1)OCI_OTYPE_NAME, (ub1)1, (ub1)OCI_PTYPE_VIEW, dschp, status);
       if (status != OCI_SUCCESS) {
 	OCIHandleFree_log_stat(dschp, OCI_HTYPE_DESCRIBE, status);
 	return oci_error(sth, errhp, status, "OCIDescribeAny(view)/LOB refetch");
@@ -1385,7 +1674,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 	return oci_error(sth, errhp, status, "OCIDescribeAny/OCIAttrGet/LOB refetch");
     }
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP, "       lob refetch from table %s, %d columns:\n",
+	PerlIO_printf(DBILOGFP, "       lob refetch from table %s, %d columns:\n",
 	    tablename, numcols);
 
     for (i = 1; i <= (long)numcols; i++) {
@@ -1406,7 +1695,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
         if (status)
                 break;
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP, "       lob refetch table col %d: '%.*s' otype %d\n",
+	    PerlIO_printf(DBILOGFP, "       lob refetch table col %d: '%.*s' otype %d\n",
 		(int)i, (int)col_name_len,col_name, col_dbtype);
 	if (col_dbtype != SQLT_CLOB && col_dbtype != SQLT_BLOB)
 	    continue;
@@ -1414,6 +1703,9 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 	    lob_cols_hv = newHV();
 	sv = newSViv(col_dbtype);
 	(void)sv_setpvn(sv, col_name, col_name_len);
+#ifdef UTF8_SUPPORT
+	DBD_SET_UTF8(sv);
+#endif
 	(void)SvIOK_on(sv);   /* what a wonderful hack! */
 	hv_store(lob_cols_hv, col_name,col_name_len, sv,0);
     }
@@ -1452,8 +1744,9 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 	while( (sv = hv_iternextsv(lob_cols_hv, &p, &i)) != NULL ) {
 	    char sql_field[200];
 	    if (phs->ora_field) {	/* must match this phs by field name	*/
+		char *ora_field_name = SvPV(phs->ora_field,na);
 		if (SvCUR(phs->ora_field) != SvCUR(sv)
-		|| ibcmp( SvPV(phs->ora_field,na), SvPV(sv,na), SvCUR(sv) ) )
+		|| ibcmp(ora_field_name, SvPV(sv,na), SvCUR(sv) ) )
 		    continue;
 	    }
 	    else			/* basic dumb match by type		*/
@@ -1467,9 +1760,9 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 		    if (phs->ftype != SvIV(sv_other))
 			continue;
 		    if (DBIS->debug >= 3)
-			fprintf(DBILOGFP,
+			PerlIO_printf(DBILOGFP,
 			"       both %s and %s have type %d - ambiguous\n",
-				SvPV(sv,na), SvPV(sv_other,na), (int)SvIV(sv_other));
+				neatsvpv(sv,0), neatsvpv(sv_other,0), (int)SvIV(sv_other));
 		    Safefree(lr);
 		    return oci_error(sth, errhp, OCI_ERROR,
 			"Need bind_param(..., { ora_field=>... }) attribute to identify table LOB field names");
@@ -1480,7 +1773,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 		(SvCUR(sql_select)>7)?", ":"", p, &phs->name[1]);
 	    sv_catpv(sql_select, sql_field);
 	    if (DBIS->debug >= 3)
-		fprintf(DBILOGFP,
+		PerlIO_printf(DBILOGFP,
 		"       lob refetch %s param: otype %d, matched field '%s' %s(%s)\n",
 		    phs->name, phs->ftype, p,
 		    (phs->ora_field) ? "by name " : "by type ", sql_field);
@@ -1497,7 +1790,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 	if (!matched) {
 	    ++unmatched_params;
 	    if (DBIS->debug >= 3)
-		fprintf(DBILOGFP,
+		PerlIO_printf(DBILOGFP,
 		    "       lob refetch %s param: otype %d, UNMATCHED\n",
 		    phs->name, phs->ftype);
 	}
@@ -1512,7 +1805,7 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
     sv_catpv(sql_select, tablename);
     sv_catpv(sql_select, " where rowid = :rid for update"); /* get row with lock */
     if (DBIS->debug >= 3)
-	fprintf(DBILOGFP,
+	PerlIO_printf(DBILOGFP,
 	    "       lob refetch sql: %s\n", SvPVX(sql_select));
     lr->sql_select = sql_select;
 
@@ -1549,11 +1842,11 @@ init_lob_refetch(SV *sth, imp_sth_t *imp_sth)
 	phs = (phs_t*)(void*)SvPVX(*phs_svp);
 	fbh->special = phs;
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       lob refetch %d for '%s' param: ftype %d setup\n",
 		(int)i+1,fbh->name, fbh->dbtype);
-	OCIDefineByPos_log_stat(lr->stmthp, &defnp, errhp, i+1,
-			 &fbh->desc_h, -1, fbh->ftype, 0,0,0, OCI_DEFAULT, status);
+	OCIDefineByPos_log_stat(lr->stmthp, &defnp, errhp, (ub4)i+1,
+			 &fbh->desc_h, -1, (ub2)fbh->ftype, 0,0,0, OCI_DEFAULT, status);
 	if (status != OCI_SUCCESS)
 	    return oci_error(sth, errhp, status, "OCIDefineByPos/LOB refetch");
     }
@@ -1613,7 +1906,7 @@ post_execute_lobs(SV *sth, imp_sth_t *imp_sth, ub4 row_count)	/* XXX leaks handl
 	    OCILobTrim_log_stat(imp_sth->svchp, errhp, fbh->desc_h, 0, status);
 	}
 	if (DBIS->debug >= 3)
-	    fprintf(DBILOGFP,
+	    PerlIO_printf(DBILOGFP,
 		"       lob refetch %d for '%s' param: ftype %d, len %ld: %s %s\n",
 		i+1,fbh->name, fbh->dbtype, ul_t(amtp),
 		(amtp > 0) ? "LobWrite" : "LobTrim", oci_status_name(status));
